
# predicted values for gaussian variables.
gaussian.prediction = function(node, fitted, data, debug = FALSE) {

  parents = fitted[[node]]$parents

  if (debug)
    cat("* predicting values for node ", node, ".\n", sep = "")

  if (length(parents) == 0) {

    .Call(call_gpred,
          fitted = fitted[[node]],
          ndata = nrow(data),
          debug = debug)

  }#THEN
  else {

    .Call(call_cgpred,
          fitted = fitted[[node]],
          parents = .data.frame.column(data, parents, drop = FALSE),
          debug = debug)

  }#ELSE

}#GAUSSIAN.PREDICTION

# predicted values for discrete networks.
discrete.prediction = function(node, fitted, data, prob = FALSE, debug = FALSE) {

  parents = fitted[[node]]$parents

  if (debug)
    cat("* predicting values for node ", node, ".\n", sep = "")

  if (length(parents) == 0) {

    .Call(call_dpred,
          fitted = fitted[[node]],
          ndata = nrow(data),
          prob = prob,
          debug = debug)

  }#THEN
  else {

    # if there is only one parent, get it easy.
    if (length(parents) == 1)
      config = .data.frame.column(data, parents)
    else
      config = configurations(.data.frame.column(data, parents), factor = FALSE)

    .Call(call_cdpred,
          fitted = fitted[[node]],
          parents = config,
          prob = prob,
          debug = debug)

  }#ELSE

}#DISCRETE.PREDICTION

# predicted values for conditional Gaussian networks.
mixedcg.prediction = function(node, fitted, data, debug = FALSE) {

  type = class(fitted[[node]])

  if (type == "bn.fit.dnode")
    discrete.prediction(node = node, fitted = fitted, data = data, debug = debug)
  else if (type == "bn.fit.gnode")
    gaussian.prediction(node = node, fitted = fitted, data = data, debug = debug)
  else {

    parents = fitted[[node]]$parents
    continuous.parents = names(which(sapply(parents, function(x) is.numeric(data[, x]))))
    discrete.parents = setdiff(parents, continuous.parents)

    # if there is only one parent, get it easy.
    if (length(discrete.parents) == 1)
      config = .data.frame.column(data, discrete.parents)
    else
      config = configurations(.data.frame.column(data, discrete.parents), factor = FALSE)

    .Call(call_ccgpred,
          fitted = fitted[[node]],
          configurations = config,
          parents = .data.frame.column(data, continuous.parents, drop = FALSE),
          debug = debug)

  }#ELSE

}#MIXEDCG.PREDICTION

# Naive Bayes and Tree-Augmented naive Bayes classifiers for discrete networks.
naive.classifier = function(training, fitted, prior, data, prob = FALSE,
    debug = FALSE) {

  # get the labels of the explanatory variables.
  nodes = names(fitted)
  # get the parents of each node, disregarding the training node.
  parents = sapply(fitted, function(x) {
    p = x$parents; return(p[p != training])
  })

  if (debug)
    cat("* predicting values for node ", training, ".\n", sep = "")

  .Call(call_naivepred,
        fitted = fitted,
        data = .data.frame.column(data, nodes, drop = FALSE),
        parents = match(parents, nodes),
        training = which(nodes == training),
        prior = prior,
        prob = prob,
        debug = debug)

}#NAIVE.CLASSIFIER

# maximum a posteriori predictions.
map.prediction = function(node, fitted, data, n, from, prob = FALSE,
    debug = FALSE) {

  .Call(call_mappred,
        node = node,
        fitted = fitted,
        data = data,
        n = as.integer(n),
        from = from,
        prob = prob,
        debug = debug)

}#MAP.PREDICTION

# prediction using exact inference in discrete networks.
exact.discrete.prediction = function(node, fitted, data, from, prob = FALSE,
    debug = FALSE) {

  # check whether gRain is loaded.
  check.and.load.package("gRain")

  # create the juntion tree.
  jtree = as.grain(fitted)

  # if the query generates a small conditional distribution (up to 50MB in
  # memory), materialize it and reuse the code that predicts from the parents;
  # otherwise, iterate over the observations.
  if (prod(sapply(data, nlevels)) <= 50 * 1024^2 / 8) {

    # generate the conditional probability table of the target variable given
    # the predictors...
    cpt = gRain::querygrain(jtree, nodes = c(node, from), type = "conditional")
    # ... embed it in a minimal bn.fit object...
    fitted.node = structure(list(node = node, parents = from, prob = cpt),
                            class = class(fitted[[node]]))
    fitted.network = structure(list(fitted.node), names = node)
    # ... and perform prediction treating predictors as parents.
    predval = discrete.prediction(node = node, fitted = fitted.network,
                data = data, prob = prob, debug = debug)

  }#THEN
  else {

    # the predicted values are a factor with the target node's levels.
    predval = factor(rep(NA, nrow(data)), levels = levels(data[, node]))
    # the prediction probabilities, if we are to return them.
    if (prob)
      classprob = matrix(NA, nrow = nlevels(predval), ncol = nrow(data))

    # for each observation...
    for (i in seq(nrow(data))) {

      # ... set the predictors as evidence...
      jpred = gRain::setEvidence(jtree, nodes = from,
                states = sapply(data[i, from, drop = FALSE], as.character))
      # ... get the probability distribution of the node being predicted...
      predprob = gRain::querygrain(jpred, nodes = node)[[node]]
      # ... and choose the level with the highest probability, breaking ties
      # randomly.
      all.maxima = names(predprob[predprob == max(predprob)])

      if (length(all.maxima) == 1)
        predval[i] = all.maxima
      else
        predval[i] = sample(all.maxima, size = 1)

      # also save the prediction probabilities.
      if (prob)
        classprob[, i] = predprob

      if (debug) {

        cat("  > prediction for observation ", i, " is '",
          as.character(predval[i]), "' with probabilities:\n", sep = "")
        cat("  ", predprob, "\n", sep = "  ")

      }#THEN

    }#FOR

    # attach the prediction probabilities to the return value if requested to.
    if (prob)
      attr(predval, "prob") = classprob

  }#ELSE

  return(predval)

}#EXACT.DISCRETE.PREDICTION

# prediction using exact inference in gaussian networks.
exact.gaussian.prediction = function(node, fitted, data, from, prob = FALSE,
    debug = FALSE) {

  # get the global distribution...
  mvn = gbn2mvnorm(fitted)
  # ... reduce it to the target variable and the predictors...
  mu = mvn$mu[c(node, from)]
  sigma = mvn$sigma[c(node, from), c(node, from)]
  # ... create the reduced network that will produce the predictions...
  prediction.network = model2network(paste0(
    paste0("[", node, "|", paste0(from, collapse = ":"), "]"),
    paste0("[", from, "]", collapse = "")
  ))
  prediction.network = mvnorm2gbn(prediction.network, mu = mu, sigma = sigma)
  # ... and predict.
  gaussian.prediction(node = node, fitted = prediction.network, data = data,
    debug = debug)

}#EXACT.GAUSSIAN.PREDICTION
